////////////////////////////////////////////////////////////////////
//                                                                //
//              Copyright (c) 1992   Namir Clement Shammas        //
//                                                                //
//  Version:  1.0.0                                  DATE 5/5/92  //
//                                                                //
//  Purpose: implements a class of doubly-linked lists with       //
//           the following operations and features:               //
//                                                                //
//             + insert new node.                                 //
//             + search for a node with a specific occurrence     //
//               of a key.                                        //
//             + delete a node with a specific occurrence of a    //
//               key.                                             //
//             + traverse the nodes of the lists.                 //
//                                                                //
////////////////////////////////////////////////////////////////////

#include "list.h"

///////////////////////////////////////////////////////////////////////
//
template<class T>
DoubleList<T>::DoubleList()
//
// Purpose: initializes a linked list object.
//
{
    nodeCount = 0;
    nodePtr = NULL;
    headPtr = NULL;
    tailPtr = NULL;
    errMsg[0] = '\0';
}

//////////////////////////////////////////////////////////////////////
//
template<class T>
DoubleList<T>& DoubleList<T>::copy(DoubleList<T>&  dl)
//
// Purpose: copies a generic doubly-linked linked.
//
// Paramaters:
//
//    input: dl - the list to be copied.
//
{
    ListNode *p, *q, *r;

    if (headPtr)
        clear();
    nodeCount = dl.nodeCount;
    nodePtr = NULL;
    headPtr = NULL;
    tailPtr = NULL;
    errMsg[0] = '\0';
    if (!dl.headPtr)
        return *this;
    headPtr = new ListNode;
    if (!headPtr) {
        strcpy(errMsg, ALLOCATE_ERROR);
        return *this;
    }
    headPtr->listNodeData = dl.headPtr->listNodeData;
    headPtr->nextNode = NULL;
    headPtr->prevNode = NULL;
    tailPtr = headPtr;
    p = headPtr;
    r = NULL;
    q = dl.headPtr->nextNode;
    while (q) {
        r = new ListNode;
        if (!r) {
            strcpy(errMsg, ALLOCATE_ERROR);
            return *this;
        }
        r->listNodeData = q->listNodeData;
        r->nextNode = NULL;
        r->prevNode = p;
        p->nextNode = r;
        tailPtr = r;
        p = p->nextNode;
        q = q->nextNode;
    }
    return *this;
}

////////////////////////////////////////////////////////////////////
//
template<class T>
char* DoubleList<T>::getErrMsg()
//
// Purpose: returns the contents of the errMsg data member and
// assigns a null string to the errMsg member.
//
{
   return errMsg ;
}

////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::insert(T& x)
//
// Purpose: performs unordered list insertion.  each new
// element becomes the last on in the list.
//
// Paramaters:
//
//    input: x - the data item inserted in the list.
//
{
    errMsg[0] = '\0';
    if (!headPtr) {
        headPtr = new ListNode;
        if (!headPtr) {
          strcpy(errMsg, ALLOCATE_ERROR);
          return false;
        }
        headPtr->listNodeData = x;
        headPtr->nextNode = NULL;
        headPtr->prevNode = NULL;
        tailPtr = headPtr;
    }
    else {
        thisPtr = new ListNode;
        if (!thisPtr) {
          strcpy(errMsg, ALLOCATE_ERROR);
          return false;
        }
        thisPtr->listNodeData = x;
        thisPtr->prevNode = tailPtr;
        thisPtr->nextNode = NULL;
        tailPtr->nextNode = thisPtr;
        tailPtr = thisPtr;
    }
    nodeCount++;
    return true;
}

////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::insert(T& x, T& y)
//
// Purpose: performs unordered list insertion.  
// element y is next to element x in the list.
//
// Paramaters:
//
//    input: x - the data item to specify the insert position.
//    input: y - the data item to be inserted after element x.
//
{
    errMsg[0] = '\0';
    if (!headPtr) {
        strcpy(errMsg, ALLOCATE_ERROR);
        return false;
    }
    else {
        if (search(x, 1)) { // x is found  
	    nodePtr = new ListNode ;
            if (!nodePtr) {
          	strcpy(errMsg, ALLOCATE_ERROR);
          	return false;
            }
            nodePtr->listNodeData = y;
            nodePtr->prevNode = thisPtr;
            nodePtr->nextNode = thisPtr->nextNode ;
	    if (thisPtr->nextNode != NULL) { // succesor exist
		thisPtr->nextNode->prevNode = nodePtr ;
		thisPtr->nextNode = nodePtr ;
	    } 
	    else { // no successor
		thisPtr->nextNode = nodePtr ;	
                tailPtr = nodePtr;
	    }
	}
	else 
	   return false ; 
    }
    nodeCount++;
    return true;
}
///////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::search(T& x, unsigned occur)
//
// Purpose: searches for x in the unordered list starting at the
// head of the list.  The pointer to the current node is returned
// in the argument list.  The complete datum of the matching node
// is accessed by parameter x (this is relevant only for structure
// data).
//
// Paramaters:
//
//    input: x - the sought data element.
//           occur - the sought occurrence of a search key.
//
//    output: thisPtr - the pointer to the sought list element.
//
{

    occur = (occur > 0) ? occur : 1;
    thisPtr = headPtr;

    while (thisPtr && occur > 0) {
        if (x == thisPtr->listNodeData)
            occur--;
        if (occur > 0)
            thisPtr = thisPtr->nextNode;
    }

    if (occur == 0) // copy data to pointer x
        x = thisPtr->listNodeData;

    return (occur == 0) ? true : false;
}

//////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::remove(T& x, unsigned occur)
//
// Purpose: attempts to delete an item from an ordered list.
// The boolean function result confirms the deletion process.
//
// Paramaters:
//
//    input: x - the data element to be deleted from the list.
//           occur - the sought occurrence of a search key.
//
{
    ListNode *p;
    bool match;

    occur = (occur > 0) ? occur : 1;
    // search for item to delete
    match = search(x, occur);

    if (match) {
        if (nodeCount == 1) {
            delete headPtr;
            headPtr = NULL;
            tailPtr = NULL;
        }
        else if (!thisPtr->nextNode) {
            // delete the last list member
            p = thisPtr->prevNode;
            p->nextNode = NULL;
            delete tailPtr;
            // link tailPtr pointer with new last element
            tailPtr = p;
        }
        else if (!thisPtr->prevNode) {
            // delete the list headPtr
            p = thisPtr->nextNode;
            p->prevNode = NULL;
            delete headPtr;
            headPtr = p; // link headPtr with new first element
        }
        else {
            thisPtr->prevNode->nextNode = thisPtr->nextNode;
            thisPtr->nextNode->prevNode = thisPtr->prevNode;
            delete thisPtr;
        }
        nodeCount--;
    }
    return match; // return function value
}

///////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::getFirstNode(T& x)
//
// Purpose: visits the first node of the generic linked list.  The
// function returns true if the list is not empty.  Otherwise it
//  yields false.
//
// Paramaters:
//
//    output: x - the data obtained from the first list node.
//
{
    if (headPtr) {
        x = headPtr->listNodeData;
        nodePtr = headPtr;
        return true;
    }
    else
        return false;
}

/////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::getLastNode(T& x)
//
// Purpose: visits the tail of the linked list.  The function yields
// true if the list is not empty.  Otherwise, the function returns
// false.
//
// Paramaters:
//
//    output: x - the obtained from the last list node.
//
{
    if (tailPtr) {
        x = tailPtr->listNodeData;
        nodePtr = tailPtr;
        return true;
    }
    else
        return false;
}

/////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::getNextNode(T& x)
//
// Purpose: visits the next node in the generic list.  The function
// returns false if an attempt is made to read beyond the tail of
// the list.  Otherwise, the function yields true.
//
// Paramaters:
//
//    output: x - the data obtained from the next list node.
//
{
    if (nodePtr && nodePtr->nextNode) {
        nodePtr = nodePtr->nextNode;
        x = nodePtr->listNodeData;
        return true;
    }
    else
        return false;
}

/////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::getNextNode(T x, T& y)
//
// Purpose: visits the next node of x in the generic list.  The function
// returns false if an attempt is made to read beyond the tail of
// the list.  Otherwise, the function yields true.
//
// Paramaters:
//
//    input: x - the node to be checked
//    output: y - the next node of x 
//	      
{
    if (search(x, 1)) { // x exist
	if (thisPtr->nextNode != NULL) {
	    y = thisPtr->nextNode->listNodeData ;
	    nodePtr = thisPtr ;
	    return true ;
	}
    }
    return false ;
}
	
/////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::getPrevNode(T& x)
//
// Purpose: visits the previous node in the generic list.  The
// function returns false if an attempt is made to read beyond
// the tail of the list.  Otherwise, yields true.
//
// Paramaters:
//
//    output: x - the data obtained from the previous list node.
//
{
    if (nodePtr && nodePtr->prevNode) {
        nodePtr = nodePtr->prevNode;
        x = nodePtr->listNodeData;
        return true;
    }
    else
        return false;
}

/////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::getPrevNode(T x, T& y)
//
// Purpose: visits the previous node in the generic list.  The
// function returns false if an attempt is made to read beyond
// the tail of the list.  Otherwise, yields true.
//
// Paramaters:
//
//    output: x - the data obtained from the previous list node.
//
{
    if (search(x, 1)) { // x exist
	if (thisPtr->prevNode != NULL) {
	    y = thisPtr->prevNode->listNodeData ;
	    nodePtr = thisPtr ;
	    return true ;
	}
    }
    return false ;
}
	
//////////////////////////////////////////////////////////////////////
//
template<class T>
bool DoubleList<T>::getData(T& x, 
                               unsigned index, 
                               bool updateNodePtr)
//
// Purpose: obtains the data at node number index.  The function
// returns true if index is less than the nodeCount member.
//
// Parameters:
//
//    input: index - the index of the sought node
//           updateNodePtr - flag to update nodePtr member.
//
//    output: x - the data obtained from the targeted list node.
//
{
  ListNode *p;

  if (index >= nodeCount)
    return false;

  p = headPtr;
  // search for the node number index
  while (index > 0) {
    p = p->nextNode;
    index--;
  }
  x = p->listNodeData;
  nodePtr = (updateNodePtr) ? p : nodePtr;
  return true;
}

/////////////////////////////////////////////////////////////////////
//
template<class T>
void DoubleList<T>::clear()
//
// Purpose: clears the linked list.
//
{
    ListNode* p;

    while (headPtr) {
        p = headPtr;
        headPtr = headPtr->nextNode;
        delete p;
        nodeCount--;
    }
    tailPtr = NULL;
    nodePtr = NULL;
}

/////////////////////////////////////////////////////////////////////
//
template<class T>
bool SortedDoubleList <T>::search(T& x, unsigned occur)
//
// Purpose: searches for x in the sorted list starting at the
// head of the list.  The complete datum of the matching node
// is accessed by parameter x (this is relevant only for structured
// data).  The function returns true if the search data is located
// in the list, otherwise returns false.
//
// Paramaters:
//
//    input: x - the sought data element.
//           occur - the sought occurrence of a search key.
//
{
    errMsg[0] = '\0';
    occur = (occur > 0) ? occur : 1;
    thisPtr = headPtr;

    while (thisPtr && occur > 0) {
        // element might be in the list ?
        if (x <= thisPtr->listNodeData)
            occur--;
        if (occur > 0)
            thisPtr = thisPtr->nextNode;
    }

/* this statement is a bug in the original version
    if (x == thisPtr->listNodeData)
*/
    if (occur == 0)
        // copy data to x
        x = thisPtr->listNodeData;

    // return function value
/* a bug 
    return (x == thisPtr->listNodeData) ? true : false;
*/
    return (occur == 0) ? true : false;

}

/////////////////////////////////////////////////////////////////////
//
template<class T>
bool SortedDoubleList <T>::insert(T& x)
//
// Purpose: performs ordered list insertion.  The function returns
// true if the insertion has taken place.
//
// Paramaters:
//
//    input: x - the reference to the data item inserted in the list.
//
{
    ListNode *p;
    bool match;

    T buffer = x;
    match = search(buffer, 1);

    if (!thisPtr) {
        // new item inserted at the tailPtr of the list
        p = new ListNode;
        if (!p) {
            strcpy(errMsg, ALLOCATE_ERROR);
            return false;
        }
        // assign key and other members
        p->listNodeData = x;
        if (headPtr) {
            tailPtr->nextNode = p;
            p->prevNode = tailPtr;
            p->nextNode = NULL;
            tailPtr = p;
        }
        else {
            headPtr = p;
            headPtr->nextNode = NULL;
            headPtr->prevNode = NULL;
            tailPtr = p;
        }
    }
    else if (!thisPtr->prevNode) {
        // new item is the new list headPtr
        p = new ListNode;
        if (!p) {
            strcpy(errMsg, ALLOCATE_ERROR);
            return false;
        }
        // assign key and other members
        p->nextNode = headPtr;
        p->prevNode = NULL;
        headPtr->prevNode = p;
        p->listNodeData = x;
        headPtr = p; // assign new pointer to the headPtr of the list
    }
    else {
        // insert new item inside the list or at its tailPtr
        p = new ListNode;
        if (!p) {
            strcpy(errMsg, ALLOCATE_ERROR);
            return false;
        }
        // link aheadPtr with the "current" node
        p->nextNode = thisPtr;
        // link back with the "previous" node
        p->prevNode = thisPtr->prevNode;
        thisPtr->prevNode->nextNode = p;
        thisPtr->prevNode = p;
        // assign key and other members
        p->listNodeData = x;
    }
    nodeCount++;
    errMsg[0] = '\0';
    return true; // return function value
}
